//
//  EnxProviderDelegates.swift
//  Enx_CallKit_iOS
//
//  Created by jaykumar on 26/05/22.
//

import UIKit
import CallKit
import Foundation
import AVFoundation

@objc
public protocol EnxCallStateDelegate:AnyObject {
    func callAnswer()
    func callReject()
    func callTimeOut()
    func callEnd()
    func callHold()
}

class EnxProviderDelegates: NSObject {
    weak var delegate: EnxCallStateDelegate?
    var callManager : EnxCallManager
    private let provider : CXProvider
    var outGoingCall : EnxCalls?
    var answerCall: EnxCalls?
    var enxManager = EnxGlobalManager.enxGlobalManager
    var callTimer : Timer!
    var isCallReceived : Bool = false
   init(callManager : EnxCallManager, delegates :EnxCallStateDelegate?) {
       self.delegate = delegates
        self.callManager = callManager
       provider = CXProvider(configuration: type(of: self).providerConfiguration)
       super.init()
       provider.setDelegate(self, queue: nil)
    }
    static var providerConfiguration : CXProviderConfiguration{
        let providerConfiguration = CXProviderConfiguration(localizedName:"EnxCall")
        providerConfiguration.supportsVideo = true
        providerConfiguration.maximumCallsPerCallGroup = 1
        providerConfiguration.supportedHandleTypes = [.phoneNumber]
        return providerConfiguration;
    }
    
    //MARK : Incoming Calls events
    func reportIncomingCalls(uuid: UUID , handle : String , hasVideo :Bool = true , completion: ((NSError?) -> Void)? = nil){
        let update = CXCallUpdate()
        update.remoteHandle = CXHandle(type: .phoneNumber, value: handle)
        update.hasVideo = hasVideo
        provider.reportNewIncomingCall(with: uuid, update: update) { [self] error in
            if error == nil {
                let call = EnxCalls(uuid: uuid)
                call.handle = handle
                self.callManager.addCall(call)
                isCallReceived = false
                self.callTimer = Timer.scheduledTimer(timeInterval: 45, target: self, selector: #selector(timeOutCall), userInfo: nil, repeats: false)
            }
            completion?(error as NSError?)
        }
    }
    func reportEndCall(){
        guard enxManager.getUUID() != nil else {
            return
        }
        provider.reportCall(with:enxManager.getUUID()!, endedAt: Date(), reason: .remoteEnded)
    }
    func senddefaultAudioInterruptionNofificationToStartAudioResource(){
        var userInfo : [AnyHashable : Any] = [:]
        let intrepEndeRaw = AVAudioSession.InterruptionType.ended.rawValue
        userInfo[AVAudioSessionInterruptionTypeKey] = intrepEndeRaw
        NotificationCenter.default.post(name: AVAudioSession.interruptionNotification, object: self, userInfo: userInfo)
    }
    func configurAudioSession(){
        let session = AVAudioSession.sharedInstance()
        do{
            try session.setCategory(AVAudioSession.Category.playAndRecord, mode: .default)
            try session.setActive(true, options: .notifyOthersOnDeactivation)
            try session.setActive(true)
            try session.setMode(AVAudioSession.Mode.voiceChat)
            try session.setPreferredSampleRate(44100.0)
            try session.setPreferredIOBufferDuration(0.005)
        }catch{
            print(error)
        }
    }
    @objc func timeOutCall(){
        delegate?.callTimeOut()
        cleanTimerInstance()
        provider.reportCall(with:enxManager.getUUID()!, endedAt: Date(), reason: .unanswered)
        enxManager.cleanAllInstance()
     }
    func cleanTimerInstance(){
        if(callTimer != nil){
            callTimer.invalidate()
            callTimer = nil
        }
    }
}
// MARK : CXProviderDelegate
extension EnxProviderDelegates : CXProviderDelegate{
    func providerDidReset(_ provider: CXProvider) {
        for call in callManager.calls{
            call.endCall()
            provider.reportCall(with: call.uuid, endedAt: Date(), reason: .remoteEnded)
        }
        callManager.removeAllCalls()
    }
    func provider(_ provider: CXProvider, perform action: CXStartCallAction) {
        let call = EnxCalls(uuid: action.callUUID, isOutgoing: true)
        call.handle = action.handle.value
        configurAudioSession()
        call.hasStartedConnectDidChange = { [weak self] in
            self?.provider.reportOutgoingCall(with: call.uuid, startedConnectingAt: call.connectData)
        }
        call.hasConnectDidChange = { [weak self] in
            self?.provider.reportOutgoingCall(with: call.uuid, startedConnectingAt: call.connectedData)
        }
        self.outGoingCall = call;
        action.fulfill()
    }
    func provider(_ provider: CXProvider, perform action: CXAnswerCallAction) {
        guard let call = callManager.callWithUUID(uuid: action.callUUID) else{
            action.fail()
            return
        }
        configurAudioSession()
        self.answerCall = call
        action.fulfill()
        isCallReceived = true
        cleanTimerInstance()
        delegate?.callAnswer()
    }
    func provider(_ provider: CXProvider, perform action: CXEndCallAction) {
        print("Receive \(#function)")
        guard let call = callManager.callWithUUID(uuid: action.callUUID) else {
            action.fail()
            return
        }
        call.endCall()
        action.fulfill()
        callManager.removeCall(call)
        provider.reportCall(with: call.uuid, endedAt: Date(), reason: .remoteEnded)
        enxManager.cleanAllInstance()
        cleanTimerInstance()
        if(!isCallReceived){
            delegate?.callReject()
        }
        else{
            delegate?.callEnd()
        }
    }
    func provider(_ provider: CXProvider, perform action: CXSetHeldCallAction) {
        guard let call = callManager.callWithUUID(uuid: action.callUUID) else {
            action.fail()
            return
        }
        call.isOnHild = action.isOnHold
        call.isMuted = action.isOnHold
        action.fulfill()
        delegate?.callHold()
    }
    func provider(_ provider: CXProvider, perform action: CXSetMutedCallAction) {
        guard let call = callManager.callWithUUID(uuid: action.callUUID) else {
            action.fail()
            return
        }
        call.isMuted = action.isMuted
        action.fulfill()
    }
    func provider(_ provider: CXProvider, timedOutPerforming action: CXAction) {
        print("Timed out Action")
        //
    }
    func provider(_ provider: CXProvider, didActivate audioSession: AVAudioSession) {
        print("Receive \(#function)")
        if(answerCall?.hasConnected ?? false){
            senddefaultAudioInterruptionNofificationToStartAudioResource()
            return
        }
        if(outGoingCall?.hasConnected ?? false){
            senddefaultAudioInterruptionNofificationToStartAudioResource()
            return
        }
        outGoingCall?.startCall(withAudioSession: audioSession) {success in
            if success {
                self.callManager.addCall(self.outGoingCall!)
                self.outGoingCall?.startAudio()
            }
        }
        answerCall?.ansCall(withAudioSession: audioSession) { success in
            if success{
                self.answerCall?.startAudio()
            }
        }
    }
    func provider(_ provider: CXProvider, didDeactivate audioSession: AVAudioSession) {
        print("Deactivate \(#function)")
        if outGoingCall?.isOnHild ?? false || answerCall?.isOnHild ?? false{
            print("Call is on hold")
            return
        }
        outGoingCall?.endCall()
        if(outGoingCall != nil){
            outGoingCall = nil
        }
        answerCall?.endCall()
        if(answerCall != nil){
            answerCall = nil
        }
        callManager.removeAllCalls()
    }
    
}
